@@ -5,6 +5,9 @@ Classes
 FMin
     Minimize an objective function to optimize a set of hyperparameters.
 """
+import pprint
+
+import matplotlib.pyplot as plt
 import numpy as np
 import scipy.stats
 
@@ -14,6 +17,7 @@ from pyrameter.domains.joint import JointDomain
 import pyrameter.methods
 from pyrameter.searchspace import SearchSpace
 from pyrameter.specification import Specification
+from pyrameter.trial import TrialStatus
 
 
 class FMin(object):
@@ -55,8 +59,9 @@ class FMin(object):
         self.searchspaces = [SearchSpace(ss, exp_key=exp_key)
                              for ss in self.spec.split()]
         self.trials = {}
+        self.active = [ss for ss in self.searchspaces]
 
-        self.max_evals = max_evals
+        self.max_evals = max_evals if max_evals is not None else np.inf
 
         if isinstance(backend, str):
             if '.json' in backend:
@@ -100,35 +105,45 @@ class FMin(object):
         if ssid is None:
             if self._did_sort:
                 probs = scipy.stats.planck.pmf(
-                    range(len(self.searchspaces)), 0.5)
+                    range(len(self.active)), 0.5)
             else:
-                probs = np.ones(len(self.searchspaces))
+                probs = np.ones(len(self.active))
             probs /= probs.sum()
-            idx = np.random.choice(np.arange(len(self.searchspaces)), p=probs)
+            idx = np.random.choice(np.arange(len(self.active)), p=probs)
 
-            while idx < len(self.searchspaces) and self.searchspaces[idx].done:
+            while idx < len(self.active) and self.active[idx].done:
                 idx += 1
 
             try:
-                trial = self.searchspaces[idx](method=self.method)
+                ss = self.active[idx]
+                if not ss.done:
+                    trial = ss(method=self.method)
+                else:
+                    trial = None
             except IndexError:
+                ss = None
                 trial = None
         else:
-            searchspace = [ss for ss in self.searchspaces if ss.id == ssid][0]
+            ss = [ss for ss in self.active if ss.id == ssid][0]
 
             if not ss.done:
-                trial = searchspace(method=self.method)
+                trial = ss(method=self.method)
             else:
                 trial = None
 
         if trial is not None:
             self.trials[trial.id] = trial
+        if ss and len(ss.trials) >= self.max_evals:
+            ss.done = True
         return trial
 
     def load(self):
         """Load experiment state from the backend."""
         if self.backend is not None:
             self.searchspaces = self.backend.load()
+            for s in self.searchspaces:
+                for t in s.trials:
+                    self.trials[t.id] = t
 
     def optimum(self):
         """Retrieve the optimal observed set of hyperparameter values."""
@@ -139,6 +154,29 @@ class FMin(object):
                 best = candidate
         return best
 
+    def plot_objective(self, show=True, save=False, filename=None):
+        for ss in self.searchspaces:
+            objs = list(filter(lambda x: x.status == TrialStatus.DONE, ss.trials))
+            objs = sorted(objs, key=lambda x: x.id)
+            objs = np.array(list(map(lambda x: x.objective, objs)))
+
+            plt.plot(np.arange(objs.shape[0]), objs, label='Space f{ss.id}')
+        
+        plt.grid(which='both')
+        plt.xlabel('Trial')
+        plt.ylabel('Objective')
+        plt.title('Search Overview')
+        plt.gcf().set_size_inches(10, 10)
+        
+        if save:
+            if filename is None:
+                filename = f'{self.exp_key}_performance.png'
+            plt.savefig(filename)
+
+        if show:
+            plt.show()
+
+
     def register_result(self, ssid, trial_id, objective=None, results=None,
                         errmsg=None):
         """Register a result in the proper trial.
@@ -157,10 +195,10 @@ class FMin(object):
         errmsg : str
             Error message output by the trial if it failed.
         """
-        searchspace = [ss for ss in self.searchspaces if str(ss.id) == ssid][0]
+        ss = [ss for ss in self.searchspaces if str(ss.id) == str(ssid)][0]
 
         if not isinstance(trial_id, list):
-            trial = [t for t in searchspace.trials if str(t.id) == str(trial_id)][0]
+            trial = [t for t in ss.trials if str(t.id) == str(trial_id)][0]
             trial.objective = objective
             trial.results = results
             trial.errmsg = errmsg
@@ -169,15 +207,16 @@ class FMin(object):
                 self.trials[trial.id] = trial
             trial.submissions += 1
 
-            if searchspace.complexity == 1:
-                n_done = sum([1 for t in searchspace.trials
+            if ss.complexity == 1:
+                n_done = sum([1 for t in ss.trials
                               if t.status.value == 3])
                 if n_done > self.max_evals:
-                    searchspace.done = True
+                    ss.done = True
+                    self.active.remove(ss)
         else:
             hyperparameters = []
             for i, tid in enumerate(trial_id):
-                trial = [t for t in searchspace.trials if str(t.id) == str(tid)][0]
+                trial = [t for t in ss.trials if str(t.id) == str(tid)][0]
                 trial.objective = objective[i]
                 trial.results = results[i]
                 trial.errmsg = errmsg
@@ -186,11 +225,12 @@ class FMin(object):
                     self.trials[trial.id] = trial
                 trial.submissions += 1
 
-                if searchspace.complexity == 1:
-                    n_done = sum([1 for t in searchspace.trials
+                if ss.complexity == 1:
+                    n_done = sum([1 for t in ss.trials
                                   if t.status.value == 3])
                     if n_done > self.max_evals:
-                        searchspace.done = True
+                        ss.done = True
+                        self.active.remove(ss)
 
         return trial.submissions, hyperparameters
 
@@ -227,6 +267,41 @@ class FMin(object):
         if use_complexity or use_uncertainty:
             self.searchspaces.sort(key=lambda x: x.rank)
             self._did_sort = True
+    
+    def to_dataframes(self):
+        dfs = [ss.to_dataframe() for ss in self.searchspaces]
+        return dfs
+
+    def summary(self):
+        total = len(self.trials)
+        success = sum([1 for v in self.trials.values() if v.status == TrialStatus.DONE])
+        pending = sum([1 for v in self.trials.values() if v.status in [TrialStatus.INIT, TrialStatus.READY]])
+        error = sum([1 for v in self.trials.values() if v.status == TrialStatus.ERROR])
+
+        opt = self.optimum()
+
+        print('\n')
+        print(f'Experiment {self.exp_key}')
+        print('-------------------------------------------------------------------')
+        print(f'{total} Trials\t{pending} Pending\t{success} Successes\t{error} Errors')
+        print('-------------------------------------------------------------------')
+        print('\n')
+        if success > 0:
+            print('Optimal')
+            print('-------------------------------------------------------------------')
+            print(f'Trial {opt.id}\t{opt.objective} Loss')
+            print('\n')
+            print('with hyperparameters')
+            print('\n')
+            pprint.pprint(opt.parameter_dict)
+            print('\n')
+            print('and additional results')
+            print('\n')
+            pprint.pprint(opt.results)
+            print('-------------------------------------------------------------------')
+            print('\n')
+        else:
+            print('No completed trials found. Check for issues and re-run the search.')
 
     @property
     def trial_count(self):
