@@ -11,15 +11,18 @@ import functools
 import itertools
 from multiprocessing.pool import ThreadPool
 import operator
+import os
 import warnings
 
 import numpy as np
+import pandas as pd
 from sklearn.gaussian_process import GaussianProcessRegressor
 from sklearn.gaussian_process.kernels import RBF
 
 from pyrameter.domains.base import Domain
+from pyrameter.domains.linked import DependentDomain
 from pyrameter.methods.random import random_search
-from pyrameter.trial import Trial
+from pyrameter.trial import Trial, TrialStatus
 
 
 class SearchSpaceMeta(type):
@@ -48,8 +51,17 @@ class SearchSpace(object, metaclass=SearchSpaceMeta):
         self._uncertainty = None
 
         self.domains = domains if domains is not None else []
+        for d1 in self.domains:
+            if isinstance(d1, DependentDomain):
+                d1_prefix = os.path.splitext(d1.name)[0]
+                d2_name = os.path.splitext(d1.domain.name)[0]
+                name = f'{d1_prefix}.{d2_name}'
+                for d2 in self.domains:
+                    if name == d2.name:
+                        d1.domain = d2 
         self.domains.sort()
 
+
         self.done = False
 
     def __call__(self, method=None, to_dict=False):
@@ -71,6 +83,8 @@ class SearchSpace(object, metaclass=SearchSpaceMeta):
         if method is None:
             method = random_search
         hyperparameters = method(self)
+        for i, d in enumerate(self.domains):
+            d.current = hyperparameters[i]
         trial = Trial(self, hyperparameters=hyperparameters)
         self.trials.append(trial)
         return trial.parameter_dict if to_dict else trial
@@ -79,6 +93,9 @@ class SearchSpace(object, metaclass=SearchSpaceMeta):
         return len(self.domains) == len(other.domains) and \
             all(map(lambda x: x[0] == x[1], zip(self.domains, other.domains)))
 
+    def __deepcopy__(self, memo):
+        return super().__deepcopy__(memo)
+
     @property
     def complexity(self):
         """Estimate the relative combinatorial complexity of this search space.
@@ -133,7 +150,7 @@ class SearchSpace(object, metaclass=SearchSpaceMeta):
         hyperparameters : list
             List of hyperparameters in order of domain name.
         """
-        return [d.generate() for d in self.domains]
+        return [d() for d in self.domains]
 
     def optimum(self, mode='min'):
         """Get the trial with the optimal performance.
@@ -166,19 +183,55 @@ class SearchSpace(object, metaclass=SearchSpaceMeta):
             of domain name, with the value of the objective as the final entry
             in the row. If no trials have been conducted, returns ``None``.
         """
+        completed = [t for t in self.trials if t.status == TrialStatus.DONE]
         if len(self.trials) > 0:
-            out = np.zeros((len(self.trials), len(self.domains) + 1),
+            out = np.zeros((len(completed), len(self.domains) + 1),
                   dtype=np.float32)
 
-            for i, result in enumerate(self.trials):
-                vec = [self.domains[j].map_to_domain(result.hyperparameters[j])
-                       for j in range(len(self.domains))]
+            for i, result in enumerate(completed):
+                vec = [float(self.domains[j].map_to_domain(result.hyperparameters[j]))
+                    for j in range(len(self.domains))]
                 vec.append(result.objective)
                 out[i] += np.asarray(vec)
         else:
             out = None
         return out
 
+    def to_dataframe(self):
+        """Convert the trials in this search space into a Pandas dataframe.
+
+        Returns
+        -------
+        df : `pandas.DataFrame`
+            A DataFrame object with rows corresponding to trials in this
+            SearchSpace. Rows include the SearchSpace id, all hyperparameters,
+            and all recorded results.
+        """
+        df_dict = {'id': [], 'index': [], 'objective': []}
+        for i, trial in enumerate(self.trials):
+            if trial.status == TrialStatus.DONE:
+                df_dict['id'].append(trial.id)
+                df_dict['index'].append(i)
+                df_dict['objective'].append(trial.objective)
+                for j, domain in enumerate(self.domains):
+                    if domain.name not in df_dict:
+                        df_dict[domain.name] = []
+                    df_dict[domain.name].append(trial.hyperparameters[j])
+                if trial.results is not None:
+                    for key, val in trial.flatten_results().items():
+                        if key not in df_dict:
+                            df_dict[key] = []
+                        
+                        if isinstance(val, np.floating):
+                            return float(val)
+                        elif isinstance(val, (np.integer, np.unsignedinteger)):
+                            return int(val)
+                        df_dict[key].append(val)
+        df = pd.DataFrame.from_dict(df_dict)
+        return df
+
+            
+
     def to_json(self, simplify=False):
         """Convert this search space to a JSON-compatible representation."""
         jsonified = {
@@ -240,3 +293,63 @@ class SearchSpace(object, metaclass=SearchSpaceMeta):
             self._uncertainty = 1
 
         return self._uncertainty
+
+
+class GridSearchSpace(SearchSpace):
+    """Hierarchical grid search domain organization and value generation.
+
+    To handle exhaustively searching a grid, instead of using a splitting
+    scheme domains are collected into iterators and values are generated in
+    order over all dimensions of the grid.
+    """
+    def __init__(self, domains, exp_key=''):
+        super().__init__(domains, exp_key=exp_key)
+        self._iterator = itertools.product(*self.domains)
+
+    def __call__(self, method=None, to_dict=False):
+        try:
+            hyperparameters = np.array(list(next(self._iterator)))
+            trial = Trial(self, hyperparameters=hyperparameters)
+            self.trials.append(trial)
+        except StopIteration:
+            self.done = True
+            return None
+
+    def restart(self):
+        self._iterator = itertools.product(*self.domains)
+
+
+class PopulationSearchSpace(SearchSpace):
+    """Hierarchical hyperparameter domain organization for population methods.
+
+    
+    """
+
+    def __init__(self, domains, population_size=50, exp_key=''):
+        super().__init__(domains, exp_key=exp_key)
+        self.population_size = population_size
+        self.population = []
+        self.best = None
+
+    def __call__(self, method=None, to_dict=False):
+        """Generate a new trial for this search space.
+
+        Parameters
+        ----------
+        to_dict : bool
+            Convert the hyperparameter values to a nested dictionary on return.
+
+        Returns
+        -------
+        trial : ``pyrameter.trial.Trial`` or dict
+            Trial data, including hyperparameter values and metadata for a
+            database. If ``to_dict`` is ``True``, instead return only the
+            nested dictionary of hyperparameter values matching the structure
+            of the original specification.
+        """
+        if method is None:
+            method = random_search
+        self.population = method(self)
+        trials = [Trial(self, hyperparameters=h) for h in self.population]
+        self.trials.extend(trials)
+        return [t.parameter_dict for t in trials] if to_dict else trials
\ No newline at end of file
